#   Copyright 2019 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Tests for projectq.ops._qaagate."""

from projectq.ops import _qaagate, All, H, X


def test_qaa_str():

    def func_algorithm(): All(H)

    def func_oracle(): All(X)

    gate = _qaagate.QAA(func_algorithm, func_oracle)
    assert str(gate) == "QAA(Algorithm = func_algorithm, Oracle = func_oracle)"
